from tnreason.optimization import satisfaction_counter as sc

from tnreason.logic import expression_utils as eu

from tnreason.contraction import core_contractor as coc
from tnreason.contraction import expression_evaluation as ee

import numpy as np

from tnreason.logic import coordinate_calculus as cc
from tnreason.representation import sampledf_to_cores as stoc

from tnreason.model import tensor_model as tm
from tnreason.model import formula_tensors as ft


class WeightEstimator:
    '''
    formulaDict: formulaKey: [expression, satRate, empRate, weight]
    '''

    def __init__(self, formulaList=[], startWeightsDict={}, sampleDf=None):
        ## Have weight saved in self.formulaDict and in self.formulaTensors
        self.formulaDict = {"f" + str(i): [formulaKey, 0, 0, 0] for i, formulaKey in enumerate(formulaList)}
        for key in self.formulaDict:
            if key in startWeightsDict:
                self.formulaDict[key][3] = startWeightsDict[key]
            else:
                startWeightsDict[key] = 0

        self.formulaTensors = tm.TensorRepresentation(
            {formulaKey: [self.formulaDict[formulaKey][0], self.formulaDict[formulaKey][3]] for formulaKey in
             self.formulaDict})

        if sampleDf is not None:
            self.load_sampleDf(sampleDf)
        else:
            self.dataTensor = None

    def add_formula(self, key, expression, weight=1):
        self.formulaTensors.add_expression(expression, weight, formulaKey=key)

        empRate = coc.CoreContractor(
            {**self.formulaTensors.get_cores([key], headType="truthEvaluation"),
             **self.dataTensor.get_cores()}).contract().values / self.dataTensor.dataNum
        satRate = coc.CoreContractor(
            {**self.formulaTensors.get_cores([key], headType="truthEvaluation")}
        ).contract().values / 2 ** len(eu.get_variables(expression))
        self.formulaDict[key] = [expression, satRate, empRate, weight]

    def load_sampleDf(self, sampleDf):
        self.dataTensor = ft.DataTensor(sampleDf)

    def calculate_independent_satRates(self):
        for formulaKey in self.formulaDict:
            self.formulaDict[formulaKey][1] = coc.CoreContractor(
                {**self.formulaTensors.get_cores([formulaKey], headType="truthEvaluation")
                 }).contract().values / 2 ** len(eu.get_variables(self.formulaDict[formulaKey][0]))

    def calculate_empRates(self):
        for formulaKey in self.formulaDict:
            self.formulaDict[formulaKey][2] = coc.CoreContractor(
                {**self.formulaTensors.get_cores([formulaKey], headType="truthEvaluation"),
                 **self.dataTensor.get_cores()}).contract().values / self.dataTensor.dataNum

    def independent_estimation(self, calculateEmp=True, calculateSat=True, cut=100):
        if calculateEmp:
            self.calculate_empRates()
        if calculateSat:
            self.calculate_independent_satRates()

        self.formulaTensors.update_heads({formulaKey: cutoff_weight(
            solve_rate_equation(self.formulaDict[formulaKey][1], self.formulaDict[formulaKey][2]), cut) for formulaKey
            in self.formulaDict})

        for formulaKey in self.formulaDict:
            self.formulaDict[formulaKey][3] = cutoff_weight(
                solve_rate_equation(self.formulaDict[formulaKey][1], self.formulaDict[formulaKey][2]), cut)

    def alternating_optimization(self, sweepNum):
        weightTracker = np.empty((sweepNum + 1, len(self.formulaDict)))
        self.independent_estimation()

        weightTracker[0] = [self.formulaDict[formulaKey][3] for formulaKey in self.formulaDict]
        for sweepPos in range(sweepNum):
            for i, formulaKey in enumerate(self.formulaDict):
                self.formula_optimization(formulaKey)
                weightTracker[sweepPos + 1, i] = self.formulaDict[formulaKey][3]
        return weightTracker

    def formula_optimization(self, tboFormulaKey):
        conDict = self.formulaTensors.get_cores(headType="expFactor")

        oldLength = len(conDict)
        conDict = {key: conDict[key] for key in conDict if key != tboFormulaKey + "_head"}
        newLength = len(conDict)

        if newLength != oldLength - 1:
            print(conDict.keys(), tboFormulaKey + "_head")
            raise ValueError("HeadCore of formula {} not found!".format(tboFormulaKey))

        negativeExpWeight, positiveExpWeight = coc.CoreContractor(conDict, openColors=[
            tboFormulaKey + "_" + str(self.formulaDict[tboFormulaKey][0])]).contract().values
        negPosQuotient = negativeExpWeight / positiveExpWeight
        empRate = self.formulaDict[tboFormulaKey][2]
        self.formulaDict[tboFormulaKey][3] = np.log(negPosQuotient * (empRate / (1 - empRate)))

        self.formulaTensors.update_heads({tboFormulaKey: np.log(negPosQuotient * (empRate / (1 - empRate)))})

    def get_weights(self):
        return {key: self.formulaDict[key][3] for key in self.formulaDict}

def calculate_satRate_bc(expression):
    return ee.ExpressionEvaluator(expression, initializeBasisCores=True).create_formula_factor().count_satisfaction()


def solve_rate_equation(satRate, empRate):
    return -np.log(((1 - empRate) * satRate) / (empRate * (1 - satRate)))


def regularize_empRate(empRate, regFactor):
    return regFactor * (empRate - 0.5) + 0.5


def cutoff_weight(weight, cutoff):
    if weight > cutoff:
        return cutoff
    else:
        return weight


## Old, used only in SampleBasedMLNLearner
def calculate_weight(expression, sampleDf, filterCore=None, regFactor=1, verbose=False, check=True, cut=20):
    atoms = eu.get_variables(expression)
    atomDict = {
        atom: cc.CoordinateCore(stoc.sampleDf_to_universal_core(sampleDf, [atom]).flatten(), ["j"])
        for atom in atoms
    }
    satRate = calculate_satRate(expression)

    if check:
        assert satRate == calculate_satRate_bc(expression), "Saturation Rate does not coincide with Basis Calculus!"

    empRate = regularize_empRate(calculate_empRate(expression, atomDict, filterCore), regFactor)
    weight = cutoff_weight(solve_rate_equation(satRate, empRate), cut)
    if verbose:
        print("## Calculationg the weight of {} ##".format(expression))
        print("World satisfaction rate: {}".format(satRate))
        print("Data satifcation rate: {}".format(empRate))
        print("Calculated weight: {}".format(weight))
    return empRate, satRate, weight


def calculate_empRate(expression, atomDict, filterCore=None):
    expressionCore = ee.ExpressionEvaluator(expression, atomDict=atomDict).evaluate()
    if filterCore is None:
        expressionResults = expressionCore.values.flatten()
        empSatNum = np.sum(expressionResults)
        dataNum = expressionResults.shape[0]
    else:
        expressionCore = expressionCore.compute_and(filterCore)
        empSatNum = np.sum(expressionCore.values)
        dataNum = np.sum(filterCore.values)
    return empSatNum / dataNum


def partition_function(satRate, w):
    return (satRate * np.exp(w)) / (1 - satRate + satRate * np.exp(w))


def calculate_satRate(expression):
    variables = np.unique(eu.get_variables(expression))
    modelNum = 2 ** len(variables)
    satNum = sc.count_satisfaction(expression)
    return satNum / modelNum


if __name__ == "__main__":
    from tnreason.logic import coordinate_calculus as cc

    from matplotlib import pyplot as plt

    atomDict = {
        "a": cc.CoordinateCore(np.random.binomial(n=1, p=0.8, size=(10, 7, 5)), ["l1", "y", "z"], name="a"),
        "b": cc.CoordinateCore(np.random.binomial(n=1, p=0.8, size=(10, 7, 5)), ["l2", "q", "z"], name="b"),
        "c": cc.CoordinateCore(np.random.binomial(n=1, p=0.4, size=(10, 7, 5)), ["l3", "q", "z"], name="c"),
    }

    expressionList = [["b", "and", ["a", "and", ["not", "c"]]],
                      "a",
                      ["not", ["b", "and", "c"]]]

    # expressionList = ["a", "b", "c"] # Needs to be constant in that case

    from tnreason.model import generate_test_data as gtd

    sampleDf = gtd.generate_sampleDf({
        "e1": ["a", 2],
        "e2": ["b", 2],
        "e3": ["c", 2],
    },
        100)

    estimator = WeightEstimator(expressionList, sampleDf=sampleDf)
    # estimator.add_formula("e4", ["a", "and", ["not", "b"]], 1)
    #    estimator.independent_estimation(atomDict)
    weightTracker = estimator.alternating_optimization(10)

    plt.imshow(weightTracker, cmap="coolwarm")
    plt.colorbar()
    plt.show()
    exit()
    ## Independent Calculation (Old)
    expression = ["b", "and", ["a", "and", ["not", "c"]]]

    filterCore = cc.CoordinateCore(np.random.binomial(n=1, p=0.4, size=(10, 10, 10, 7, 7, 5)),
                                   ["l1", "l2", "l3", "y", "q", "z"], name="c")

    checkEmpRate = calculate_empRate(expression, atomDict, filterCore)
    checkSatRate = calculate_satRate(expression)
    result = calculate_weight(expression, atomDict, filterCore)



import pandas as pd


def sampleDf_to_factDf(sampleDf,prefix=""):
    columns = sampleDf.columns
    df = pd.DataFrame(columns = ["subject","predicate","object"])
    index = 0
    for i, row in sampleDf.iterrows():
        for column in columns:
           triple = generate_triple(i,column,row[column],prefix = prefix)
           if triple is not None:
                    row_dict = {"subject": triple[0],
                                "predicate": triple[1],
                                "object": triple[2]
                    }
                    row_df = pd.DataFrame(row_dict, index=[index])
                    index +=1
                    df = pd.concat([df, row_df])
    return df






## From createDG:
def generate_triple(index,key,value,prefix):
    if "," in key:
        triple = generate_relation_triple(index,key,value,prefix)
    else:
        triple = generate_membership_triple(index,key,value,prefix)
    return triple

def generate_relation_triple(index,key,value,prefix):
    if value == 1:
        predpart, argpart = key.split("(")
        arg1, arg2 = argpart.split(",")

        s = prefix + arg1 + "_" + str(index)
        p = prefix + predpart
        o = prefix + arg2[:-1] + "_" + str(index)
        return s, p, o
    else:
        return None

def generate_membership_triple(index,key,value,prefix):
    if value == 1:
        predpart, argpart = key.split("(")

        s = prefix + argpart[:-1] + "_" + str(index)
        p = "rdf:type"
        o = prefix + predpart
        return s, p, o
    else:
        return None


from tnreason.model import model_visualization as mv


class LogicRepresentation:
    def __init__(self, expressionsDict):
        self.expressionsDict = expressionsDict

    def infer(self, evidenceDict, simplify=True):
        self.expressionsDict = {
            key: [replace_evidence_variables(self.expressionsDict[key][0], evidenceDict), self.expressionsDict[key][1]]
            for key in
            self.expressionsDict
        }
        if simplify:
            self.simplify()

    def forward_chaining(self, hardLogicLimit=100):
        ## Synthetic algorithm: Formulas with weights crossing the hardLogicLimit are taking for hard logical inference
        newEvidenceDict = find_new_evidence(self.expressionsDict, hardLogicLimit)
        evidenceDict = newEvidenceDict.copy()
        while not len(newEvidenceDict) == 0:
            self.infer({atomKey: bool(newEvidenceDict[atomKey] / abs(newEvidenceDict[atomKey])) for atomKey in
                        newEvidenceDict})
            newEvidenceDict = find_new_evidence(self.expressionsDict, hardLogicLimit)
            evidenceDict = {**evidenceDict, **newEvidenceDict}
        self.expressionsDict = {**self.expressionsDict,
                                **{atomKey + "_evidence": [atomKey, evidenceDict[atomKey]] for atomKey in evidenceDict}}

    def simplify(self):
        self.remove_thing_nothing()
        self.remove_double_nots()
        self.remove_doubles()
        self.beautify_weights()

    def remove_thing_nothing(self):
        newExpressionsDict = {}
        for key in self.expressionsDict:
            newExpression = reduce_thing_nothing(self.expressionsDict[key][0])
            if newExpression not in ["Thing", "Nothing"]:
                newExpressionsDict[key] = [newExpression, self.expressionsDict[key][1]]
        self.expressionsDict = newExpressionsDict

    def remove_double_nots(self):
        self.expressionsDict = {key: [reduce_double_not(self.expressionsDict[key][0]), self.expressionsDict[key][1]] for
                                key in self.expressionsDict}

    def remove_doubles(self):
        # Removes Expressions which are the same or negations of each other
        checkedKeys = []
        reducedExpressionDict = {}
        for key in self.expressionsDict:
            if key not in checkedKeys and self.expressionsDict[key][0]:
                checkedKeys.append(key)
                keyFormula, keyWeight = self.expressionsDict[key]
                for otherKey in self.expressionsDict:
                    if otherKey not in checkedKeys:
                        result = equality_contradiction_check(keyFormula, self.expressionsDict[otherKey][0])
                        if result == "equal":
                            checkedKeys.append(otherKey)
                            keyWeight += self.expressionsDict[otherKey][1]
                        elif result == "negequal":
                            checkedKeys.append(otherKey)
                            keyWeight += - self.expressionsDict[otherKey][1]
                if keyWeight != 0:
                    reducedExpressionDict[key] = [keyFormula, keyWeight]
        self.expressionsDict = reducedExpressionDict

    def beautify_weights(self):
        self.expressionsDict = {key: posify_weight(self.expressionsDict[key][0], self.expressionsDict[key][1]) for key
                                in self.expressionsDict
                                if self.expressionsDict[key][1] != 0}

    def visualize(self, evidenceDict={}, strengthMultiplier=4, strengthCutoff=10, fontsize=10, showFormula=True,
                  pos=None):
        return mv.visualize_model(self.expressionsDict,
                                  strengthMultiplier=strengthMultiplier,
                                  strengthCutoff=strengthCutoff,
                                  fontsize=fontsize,
                                  showFormula=showFormula,
                                  evidenceDict=evidenceDict,
                                  pos=pos)


def infer_expression(expression, evidenceDict):
    return reduce_thing_nothing(replace_evidence_variables(expression, evidenceDict))


def replace_evidence_variables(expression, evidenceDict):
    # print(expression, type(expression)==str, expression in list(evidenceDict.keys()))
    if type(expression) == str:
        if expression in evidenceDict.keys():
            if bool(evidenceDict[expression]) == True:
                return "Thing"
            else:
                return "Nothing"
        else:
            return expression
    elif expression[0] == "not":
        return ["not", replace_evidence_variables(expression[1], evidenceDict)]
    elif expression[1] == "and":
        return [replace_evidence_variables(expression[0], evidenceDict), "and",
                replace_evidence_variables(expression[2], evidenceDict)]


def reduce_thing_nothing(expression):
    if type(expression) == str:
        return expression
    elif expression[0] == "not":
        rightExpression = reduce_thing_nothing(expression[1])
        if rightExpression == "Thing":
            return "Nothing"
        elif rightExpression == "Nothing":
            return "Thing"
        return ["not", rightExpression]
    elif expression[1] == "and":
        leftExpression = reduce_thing_nothing(expression[0])
        rightExpression = reduce_thing_nothing(expression[2])
        if leftExpression == "Thing":
            return rightExpression
        elif rightExpression == "Thing":
            return leftExpression
        elif leftExpression == "Nothing" or rightExpression == "Nothing":
            return "Nothing"
        return [leftExpression, "and", rightExpression]


def reduce_double_not(expression):
    if type(expression) == str:
        return expression
    elif expression[0] == "not":
        if expression[1][0] == "not":
            return reduce_double_not(expression[1][1])
        else:
            return ["not", reduce_double_not(expression[1])]
    elif expression[1] == "and":
        return [reduce_double_not(expression[0]), "and", reduce_double_not(expression[2])]
    else:
        raise ValueError("Expression {} not understood!".format(expression))


def equality_contradiction_check(expression1, expression2):
    if equality_check(expression1, expression2):
        return "equal"
    if len(expression1) > 1:
        if expression1[0] == "not":
            if equality_check(expression1[1], expression2):
                return "negequal"
    if len(expression2) > 1:
        if expression2[0] == "not":
            if equality_check(expression2[1], expression1):
                return "negequal"
    return "neither"


## Copied from logic/expression_generation
def equality_check(expression1, expression2):
    if type(expression1) == str:
        if type(expression2) == str:
            return expression1 == expression2
        else:
            return False
    elif expression1[0] == "not":
        if expression2[0] == "not":
            return equality_check(expression1[1], expression2[1])
        else:
            return False
    elif expression1[1] == "and":
        if expression2[1] == "and":
            return (equality_check(expression1[0], expression2[0]) and equality_check(expression1[2],
                                                                                      expression2[2])) or (
                    equality_check(expression1[0], expression2[2]) and equality_check(expression1[2],
                                                                                      expression2[0]))


def find_new_evidence(expressionsDict, hardLogicLimit=100):
    newEvidenceDict = {}
    for key in expressionsDict:
        if abs(expressionsDict[key][1]) > hardLogicLimit:
            if type(expressionsDict[key][0]):
                if expressionsDict[key][0] in newEvidenceDict:
                    newEvidenceDict[expressionsDict[key][0]] += expressionsDict[key][1]
                else:
                    newEvidenceDict[expressionsDict[key][0]] = expressionsDict[key][1]
            elif len(expressionsDict[key][0]) == 2:
                if type(expressionsDict[key][0][1]) == str:
                    if expressionsDict[key][0][1] in newEvidenceDict:
                        newEvidenceDict[expressionsDict[key][0][1]] += -expressionsDict[key][1]
                    else:
                        newEvidenceDict[expressionsDict[key][0][1]] = -expressionsDict[key][1]
    return newEvidenceDict


def posify_weight(expression, weight):
    if weight < 0:
        if type(expression) == str:
            return [["not", expression], -weight]
        elif expression[0] == "not":
            return [expression, -weight]
        else:
            return [["not", expression], -weight]
    return [expression, weight]


if __name__ == "__main__":
    redundantDict = {
        "f0": ["A1", 101.2312312],
        "f1": [["not", ["A2", "and", "A3"]], 1],
        "f2": ["A2", 90],
        "f3": [["not", "A2"], 90],
        "f4": ["Thing", 1],
        "f5": ["A3", -10]
    }

    # print(infer_expression(["not", ["A2", "and", "A3"]], {"A2":1}))

    lRep = LogicRepresentation(redundantDict)
    lRep.simplify()

    lRep.visualize(strengthCutoff=3)

    print(lRep.expressionsDict)

    lRep.forward_chaining()
    print(lRep.expressionsDict)

    lRep.infer({"A2": 1}, simplify=True)
    print(lRep.expressionsDict)

    lRep.remove_doubles()
    print(lRep.expressionsDict)

    and_expression = ['not', [['not', 'jaszczur'], 'and',
                              ['not', ['not', [['not', 'sikorka'], 'and', ['not', ['not', 'sledz']]]]]]]
    assert reduce_double_not(and_expression) == ['not',
                                                 [['not', 'jaszczur'], 'and', [['not', 'sikorka'], 'and', 'sledz']]], \
        "Generate from disjunctions or double not does not work"
    assert reduce_double_not(["not", ["not", "sledz"]]) == "sledz", "Removing double not does not work"



from tnreason.logic import expression_utils as eu
from tnreason.logic import coordinate_calculus as cc

from tnreason.model import tensor_model as tm
from tnreason.model import logic_model as lm
from tnreason.model import model_visualization as mv

import numpy as np
import pandas as pd


class SamplerBase:
    def __init__(self, expressionsDict):
        self.expressionsDict = expressionsDict
        self.atoms = eu.get_all_variables([self.expressionsDict[formulaKey][0] for formulaKey in self.expressionsDict])
        self.marginalizedDict = None

    def compute_marginalized_distributions(self):
        tensorRepresented = tm.TensorRepresentation(self.expressionsDict, headType="expFactor")
        self.marginalizedDict = {atomKey: tensorRepresented.marginalized_contraction([atomKey]).normalize()
                                 for atomKey in self.atoms}

    def create_independent_sample(self):
        if self.marginalizedDict is None:
            self.compute_marginalized_distributions()
        return {atomKey: np.random.multinomial(1, self.marginalizedDict[atomKey].values)[0] == 0 for atomKey in
                self.atoms}

    def infer_expressionsDict(self, evidenceDict={}):
        logRep = lm.LogicRepresentation(self.expressionsDict)
        logRep.infer(evidenceDict=evidenceDict, simplify=True)
        return logRep.expressionsDict

    def visualize(self, evidenceDict={}, strengthMultiplier=4, strengthCutoff=10, fontsize=10, showFormula=True,
                  pos=None, savePath=None):
        return mv.visualize_model(self.expressionsDict,
                                  strengthMultiplier=strengthMultiplier,
                                  strengthCutoff=strengthCutoff,
                                  fontsize=fontsize,
                                  showFormula=showFormula,
                                  evidenceDict=evidenceDict,
                                  pos=pos,
                                  savePath=savePath,
                                  show=False)


class GibbsSampler(SamplerBase):
    def create_sampleDf(self, sampleNum, chainLength=10, outType="int64"):
        sampleDf = pd.DataFrame(columns=self.atoms)
        for samplePos in range(sampleNum):
            sampleDf = pd.concat(
                [sampleDf, pd.DataFrame(self.gibbs_sample(chainLength=chainLength), index=[samplePos])])
        return sampleDf.astype(outType)

    def gibbs_sample(self, chainLength):
        sampleDict = self.create_independent_sample()
        for sweep in range(chainLength):
            for updateAtomKey in self.atoms:
                miniSampler = SamplerBase(self.infer_expressionsDict(
                    {atomKey: sampleDict[atomKey] for atomKey in self.atoms if atomKey != updateAtomKey}))
                if updateAtomKey not in miniSampler.atoms:
                    miniSampler = SamplerBase({updateAtomKey: [str(updateAtomKey), 0]})
                sampleDict[updateAtomKey] = miniSampler.create_independent_sample()[updateAtomKey]
        return sampleDict


class ExactSampler:
    def __init__(self, expressionsDict, margAtoms=None):
        self.expressionsDict = expressionsDict
        self.atoms = eu.get_all_variables([self.expressionsDict[formulaKey][0] for formulaKey in self.expressionsDict])

        if margAtoms is None:
            margAtoms = self.atoms
        self.compute_distributionCore(margAtoms)

    def compute_distributionCore(self, margAtoms):
        tensorRepresented = tm.TensorRepresentation(self.expressionsDict, headType="expFactor")
        self.distributionCore = tensorRepresented.marginalized_contraction(margAtoms).normalize()

    def compute_superposedSampleCore(self, sampleNum):
        return cc.CoordinateCore(np.random.multinomial(sampleNum, self.distributionCore.values.flatten()).reshape(
            self.distributionCore.values.shape), self.distributionCore.colors)


if __name__ == "__main__":
    learnedFormulaDict = {
        "f0": ["A1", 1],
        "f1": [["not", ["A2", "and", "A3"]], 1],
        "f2": ["A2", -1]
    }

    sampler = GibbsSampler(learnedFormulaDict)
    pos = None
    for rep in range(10):
        sample = sampler.gibbs_sample(chainLength=10)
        pos = sampler.visualize(evidenceDict=sample, pos=pos, savePath=None)#"./demonstration/visualizations/gibbs{}.png".format(rep))

    sampler.compute_marginalized_distributions()
    print(sampler.create_sampleDf(100, 20))

    exit()

    exactSampler = ExactSampler(learnedFormulaDict)
    sampleNums = [1, 10, 100, 1000, 10000]
    for sampleNum in sampleNums:
        print(sampleNum, np.linalg.norm(
            exactSampler.distributionCore.values - exactSampler.compute_superposedSampleCore(
                sampleNum).values / sampleNum))



import numpy as np
import time

import tnreason.logic.coordinate_calculus as cc

def identify_relevant_columns(columns,argument):
    relcolumns = []
    for column in columns:
        if column.split("(")[1][:-1] == argument:
            relcolumns.append(column)
    return relcolumns

def sampleDf_to_class_values(sampleDf,
                           individual):
    startTime = time.time()
    dataNum = sampleDf.values.shape[0]

    relColumns = identify_relevant_columns(sampleDf.columns,individual)
    coreValues = np.zeros((dataNum,len(relColumns)))

    for i, row in sampleDf.iterrows():
        for column in relColumns:
            if row[column] == 1:
                coreValues[i,relColumns.index(column)] = 1
    endTime = time.time()
    return coreValues, relColumns, endTime-startTime

def sampleDf_to_relation_values(sampleDf,
                              individual1,
                              individual2):
    startTime = time.time()
    dataNum = sampleDf.values.shape[0]

    relColumns = identify_relevant_columns(sampleDf.columns, individual1+","+individual2)
    coreValues = np.zeros((dataNum, len(relColumns), dataNum))

    for i, row in sampleDf.iterrows():
        for column in relColumns:
            if row[column] == 1:
                coreValues[i, relColumns.index(column), i] = 1
    endTime = time.time()
    return coreValues, relColumns, endTime - startTime

def sampleDf_to_universal_core(sampleDf,candidates):
    return sampleDf[candidates].values


def create_fixedCore(sampleDf, candidates, coreColors, coreName):
    return cc.CoordinateCore(sampleDf_to_universal_core(sampleDf, candidates), coreColors, coreName)


import numpy as np
from queue import PriorityQueue


class ContractionOptimizerBase:
    def __init__(self, coreColorDict, colorDimDict, coreList=None, globallyOpenColors=[]):
        self.coreColorDict = coreColorDict  ## Colors of each core
        self.colorDimDict = colorDimDict  ## Dimension of each color
        self.globallyOpenColors = globallyOpenColors

        if coreList is None:
            self.set_random_coreList()
        else:
            self.coreList = coreList

        self.currentSet = []

    def set_coreList(self, coreList):
        self.coreList = coreList

    def set_random_coreList(self):
        self.coreList = np.array(list(self.coreColorDict.keys()))
        np.random.shuffle(self.coreList)

    def find_openColors_at(self, pos=None, coreList=None):  # Contraction Size of coreList up to position pos
        if coreList is None:
            coreList = self.coreList
        if pos is not None:
            contractedCores = coreList[:pos]
            uncontractedCores = coreList[pos:]
        else:
            contractedCores = coreList
            uncontractedCores = [coreKey for coreKey in self.coreColorDict if coreKey not in coreList]

        contractedColors = compute_colors_in_cores(contractedCores, self.coreColorDict)
        restColors = compute_colors_in_cores(uncontractedCores, self.coreColorDict)

        openColors = []
        for color in contractedColors:
            if color in restColors or color in self.globallyOpenColors:
                openColors.append(color)

        openShape = [self.colorDimDict[color] for color in openColors]
        openSize = np.prod(openShape)

        return openColors, openShape, openSize

    def evaluate_coreList(self, coreList=None):
        if coreList is None:
            coreList = self.coreList
        self.openColors = []
        self.openShapes = []
        self.openSizes = []
        for pos in range(len(coreList) + 1):
            colors, shape, size = self.find_openColors_at(pos, coreList)
            self.openColors.append(colors)
            self.openShapes.append(shape)
            self.openSizes.append(size)
        return self.openShapes


class GreedyHeuristicOptimizer(ContractionOptimizerBase):
    def compute_heuristic_at(self, pos, newColorPenalty=1, colorCloseScore=1):
        openColors, _, _ = self.find_openColors_at(pos)
        heuristic = np.ones(len(self.coreList[pos:]))

        ## Compute color close score: Globally open colors are excluded
        for color in openColors:
            colorCount = len([core for core in self.coreList[pos:] if color in self.coreColorDict[core]])
            for i, core in enumerate(self.coreList[pos:]):
                if color in self.coreColorDict[core] and color not in self.globallyOpenColors:
                    heuristic[i] = heuristic[i] * ((colorCloseScore * self.colorDimDict[color]) / colorCount)

        ## Compute new color penalty
        for i, core in enumerate(self.coreList[pos:]):
            for color in self.coreColorDict[core]:
                if color not in openColors:
                    heuristic[i] = heuristic[i] / (newColorPenalty * self.colorDimDict[color])
        return heuristic

    def optimize(self, newColorPenalty=1, colorCloseScore=1, verbose=False):
        if verbose:
            openSizes = self.evaluate_coreList()
            print(
                "Before optimization have a footprint {} using order {}.".format(np.sum(openSizes), self.coreList))
        for pos in range(len(self.coreList)):
            maxPos = np.argmax(
                self.compute_heuristic_at(pos, newColorPenalty=newColorPenalty, colorCloseScore=colorCloseScore))
            self.coreList[pos], self.coreList[pos + maxPos] = self.coreList[pos + maxPos], self.coreList[pos]

        if verbose:
            openSizes = self.evaluate_coreList()
            print("### Summary of the Greedy Heuristic Optimization ###")
            print("The solution is {}.".format(self.coreList))
            print(
                "After optimization have a footprint {} using order {}.".format(np.sum(openSizes), self.coreList))


class SimulatedAnnealingOptimizer(ContractionOptimizerBase):

    def random_modification(self, temp=1, criterion="memory"):
        modCoreList = self.coreList.copy()

        previousShapes = self.evaluate_coreList(modCoreList)
        previousScore = np.sum([np.prod(shape) for shape in previousShapes])

        modPos1 = np.random.randint(0, len(self.coreList))
        modPos2 = np.random.randint(0, len(self.coreList))

        modCoreList[modPos1], modCoreList[modPos2] = modCoreList[modPos2], modCoreList[modPos1]

        if criterion == "memory":
            newShapes = self.evaluate_coreList(modCoreList)
            newScore = np.sum([np.prod(shape) for shape in newShapes])

            acceptanceProb = 1 / (1 + np.exp((newScore - previousScore) / temp))
        else:
            raise ValueError("Criterion {} not understood in randomized coreList optimization.".format(criterion))

        accept = np.random.choice([0, 1], p=[1 - acceptanceProb, acceptanceProb])
        if accept:
            self.coreList = modCoreList
        return accept and modPos1 != modPos2

    def metropolis(self, repetitions=10, temp=1, criterion="memory", verbose=True):
        if verbose:
            firstShapes = self.evaluate_coreList()
            firstScore = np.sum([np.prod(shape) for shape in firstShapes])
            modCounter = 0
        for repetition in range(repetitions):
            modified = self.random_modification(temp=temp, criterion=criterion)
            if modified and verbose:
                modCounter += 1
        if verbose:
            print("### Summary of the Metropolis Optimization ###")
            print("The solution is {}.".format(self.coreList))
            print("Of {} modifications {} have been accepted at temperature {}.".format(repetitions, modCounter, temp))
            newShapes = self.evaluate_coreList()
            newScore = np.sum([np.prod(shape) for shape in newShapes])
            print("The criterion {} changed from {} to {}.".format(criterion, firstScore, newScore))

    def optimize(self, coolingPattern=[[10, 10], [1, 10], [0.1, 10]], verbose=True):
        for coolingEntry in coolingPattern:
            self.metropolis(repetitions=coolingEntry[1], temp=coolingEntry[0], verbose=verbose)


class DijkstraOptimizer(ContractionOptimizerBase):

    def step(self, queueEntry):
        entryCost, entryList = queueEntry
        for coreKey in self.coreColorDict:
            if coreKey not in entryList:
                addList = entryList.copy()
                addList.append(coreKey)

                if len(addList) == len(self.coreColorDict):
                    self.coreList = addList
                    return False
                else:
                    openColors, openShape, openSize = self.find_openColors_at(coreList=entryList)
                    self.frontier.put((entryCost + openSize, addList))
        return True

    def optimize(self, verbose=True):
        self.frontier = PriorityQueue()
        self.frontier.put((0, []))

        counter = 0
        searching = True
        while searching:
            searching = self.step(self.frontier.get())
            counter += 1
        if verbose:
            print("### Summary of the Dijsktra Optimization ###")
            print("The solution is {} and has been found after {} Dijkstra steps.".format(self.coreList, counter))


def compute_colors_in_cores(cores, coreColorDict):
    colors = []
    for core in cores:
        for color in coreColorDict[core]:
            if color not in colors:
                colors.append(color)
    return colors


if __name__ == "__main__":
    cCDict = {
        "C1": ["x", "y", "z", "q", "r"],
        "C2": ["x", "y", "z", "q", "r"],
        "C3": ["x"],
        "C4": ["y"],
        "C5": ["y"]
    }

    cDDict = {
        "x": 2,
        "y": 3,
        "z": 1,
        "q": 20,
        "r": 30
    }

    optim = GreedyHeuristicOptimizer(cCDict, cDDict, globallyOpenColors=["x"])
    optim.optimize(colorCloseScore=10)

    optim2 = SimulatedAnnealingOptimizer(cCDict, cDDict, globallyOpenColors=["x"])
    optim2.metropolis(repetitions=int(10))

    optim3 = DijkstraOptimizer(cCDict, cDDict, globallyOpenColors=["x"])
    optim3.optimize()



import pandas as pd
import numpy as np
import time

def pairDf_to_target_values(pairDf, individualsdict, individuals):
    startTime = time.time()

    for individual in individuals:
        if individualsdict[individual] is not None:
            pairDf = pairDf[pairDf[individual].isin(individualsdict[individual])]
        else:
            individualsdict[individual] = pairDf[individual].values
        individualsdict[individual] = np.array(individualsdict[individual])

    repTensor = np.zeros(shape=[len(individualsdict[individual]) for individual in individuals])
    for i, row in pairDf.iterrows():
        position = tuple(int(np.argwhere(individualsdict[individual]==row[individual])) for individual in individuals)
        repTensor[position] = 1
    endTime = time.time()
    return repTensor, endTime - startTime


from tnreason.logic import expression_utils as eu
from tnreason.logic import coordinate_calculus as cc

from tnreason.contraction import layers as lay

import numpy as np

def generate_negation_layer(skeletonExpression, shapesDict, sampleNum):
    variableColors = eu.get_variables(skeletonExpression)
    variableShapes = [shapesDict[color] for color in variableColors]

    directCore = cc.CoordinateCore(np.eye(sampleNum), [str(skeletonExpression[0]), str(skeletonExpression)])
    constantCore = cc.CoordinateCore(np.ones(shape=variableShapes+[sampleNum]), variableColors + [str(skeletonExpression)])

    return lay.Layer({str(skeletonExpression)+"_"+"constant": constantCore,
                      str(skeletonExpression)+"_"+"direct": directCore},
                     {str(skeletonExpression) + "_" + "constant": 1,
                      str(skeletonExpression) + "_" + "direct": -1})

def generate_conjunction_layer(skeletonExpression, sampleNum):
    values = np.empty(shape=(sampleNum, sampleNum, sampleNum))
    for i in range(sampleNum):
        values[i, i, i] = 1
    conjunctionCore = cc.CoordinateCore(values,[str(skeletonExpression[0]),str(skeletonExpression[2]),str(skeletonExpression)])
    return lay.Layer({str(skeletonExpression): conjunctionCore},
                     {str(skeletonExpression): 1})

def generate_skeletonLayerDict(skeletonExpression, shapesDict, sampleNum):
    addCoreKey = str(skeletonExpression)
    if type(skeletonExpression) == str:
        return {}
    elif skeletonExpression[0] == "not":
        layerDict = generate_skeletonLayerDict(skeletonExpression[1], shapesDict, sampleNum)
        layerDict[addCoreKey] = generate_negation_layer(skeletonExpression, shapesDict, sampleNum)
        return layerDict
    elif skeletonExpression[1] == "and":
        layerDict = {**generate_skeletonLayerDict(skeletonExpression[0], shapesDict, sampleNum),
                     **generate_skeletonLayerDict(skeletonExpression[2], shapesDict, sampleNum)}
        layerDict[addCoreKey] = generate_conjunction_layer(skeletonExpression, sampleNum)
        return layerDict

if __name__ == "__main__":
    print(generate_skeletonLayerDict(["a","and",["not","b"]], {"a": 2, "b": 10}, 4).keys())
    print(generate_skeletonLayerDict(["a","and",["not","b"]], {"a": 2, "b": 10}, 4)["['not', 'b']"].colors)


## Integrates the former separated weight and formula learning
from tnreason.optimization import alternating_mle as amle
from tnreason.optimization import weight_estimation as wees


class FormulaLearner:
    def __init__(self, learnedFormulaDict={}, sampleDf=None):
        self.learnedFormulaDict = learnedFormulaDict
        self.weightEstimator = wees.WeightEstimator(sampleDf=sampleDf)
        for formulaKey in learnedFormulaDict:
            self.weightEstimator.add_formula(formulaKey, learnedFormulaDict[formulaKey][0],
                                             learnedFormulaDict[formulaKey][1])
        self.load_sampleDf(sampleDf)

    def load_sampleDf(self, sampleDf):
        self.sampleDf = sampleDf
        self.weightEstimator.sampleDf = sampleDf

    def learn(self, skeletonExpression, candidatesDict, formulaKey=None, gdSweepNum=10):
        ## Step1: Formula Learning
        ## For each Formula Learning Step initialize an optimizer
        optimizer = amle.GradientDescentMLE(skeletonExpression, candidatesDict,
                                            learnedFormulaDict=self.learnedFormulaDict,
                                            sampleDf=self.sampleDf)
        optimizer.alternating_gradient_descent(sweepNum=gdSweepNum, stepWidth=1)
        solutionExpression = optimizer.get_solution_expression()

        if formulaKey is None:
            formulaKey = str(solutionExpression)

        ## Step2: Weight Adjusting
        self.weightEstimator.add_formula(formulaKey, solutionExpression, weight=0)
        self.weightEstimator.formula_optimization(formulaKey)
        assignedWeight = self.weightEstimator.formulaDict[formulaKey][3]
        self.learnedFormulaDict[formulaKey] = [solutionExpression, assignedWeight]

        print("The formula {} has been added with weight {}.".format(solutionExpression, assignedWeight))

    def adjust_weights(self, sweepNum=10):
        self.weightEstimator.alternating_optimization(sweepNum)
        weightDict = self.weightEstimator.get_weights()
        for key in self.learnedFormulaDict:
            self.learnedFormulaDict[key][1] = weightDict[key]

    def get_learned_formulas(self):
        return self.learnedFormulaDict
